import typing as t

import pytest

from izulu import _reraise

_SELF = object()


@pytest.mark.parametrize("reraising", (True, False))
def test_reraising_bool_learned(reraising):
    kls = type(
        "Klass", (_reraise.ReraisingMixin,), {"__reraising__": reraising}
    )

    assert kls._ReraisingMixin__reraising == reraising


@pytest.mark.parametrize(
    "reraising",
    (
        ((ZeroDivisionError, ValueError),),
        ((ValueError, TypeError),),
        ((NotImplementedError, AttributeError),),
        ((ValueError, getattr(t, "Self", _SELF)),),
        ((TypeError, "Klass"),),
    ),
)
def test_reraising_by_class(reraising):
    kls = type(
        "Klass",
        (
            Exception,
            _reraise.ReraisingMixin,
        ),
        {"__reraising__": reraising},
    )
    original, reraised = reraising[0]
    if isinstance(reraised, str) or reraised is getattr(t, "Self", _SELF):
        reraised = kls

    # test action scope
    with pytest.raises(reraised), kls.reraise():
        raise original

    # test non-action scope
    with pytest.raises(Exception), kls.reraise():
        raise Exception


@pytest.mark.parametrize(
    "reraising",
    (
        ((ZeroDivisionError, ValueError),),
        ((ValueError, TypeError),),
        ((NotImplementedError, AttributeError),),
        ((ValueError, getattr(t, "Self", _SELF)),),
        ((TypeError, "Klass"),),
    ),
)
def test_reraising_reraise_by_class(reraising):
    kls = type(
        "Klass",
        (
            Exception,
            _reraise.ReraisingMixin,
        ),
        {"__reraising__": reraising},
    )
    original, reraised = reraising[0]
    if isinstance(reraised, str) or reraised is getattr(t, "Self", _SELF):
        reraised = kls

    # test action scope
    with pytest.raises(reraised), kls.reraise():
        raise original

    # test non-action scope
    with pytest.raises(Exception), kls.reraise():
        raise Exception


# @pytest.mark.parametrize(
#     "reraising",
#     (
#         ((ZeroDivisionError, ValueError),),
#         ((ValueError, TypeError),),
#         ((Exception, AttributeError),),
#     ),
# )
# def test_reraising_reraise_by_self(reraising):
#     kls = type(
#         "Klass",
#         (Exception, _reraise.ReraisingMixin,),
#         {"__reraising__": reraising},
#     )
#     original, reraised = reraising[0]
#
#     with pytest.raises(reraised):
#         with kls.reraise():
#             raise original


# @pytest.mark.parametrize(
#     ("kls", "fields", "hints", "registered", "defaults", "consts"),
#     (
#             (
#                     errors.RootError,
#                     frozenset(),
#                     types.MappingProxyType({}),
#                     frozenset(),
#                     frozenset(),
#                     types.MappingProxyType({}),
#             ),
#             (
#                     errors.TemplateOnlyError,
#                     frozenset(("name", "age")),
#                     types.MappingProxyType({}),
#                     frozenset(("name", "age")),
#                     frozenset(),
#                     types.MappingProxyType({}),
#             ),
#             (
#                     errors.AttributesOnlyError,
#                     frozenset(),
#                     types.MappingProxyType(dict(name=str, age=int)),
#                     frozenset(("name", "age")),
#                     frozenset(),
#                     types.MappingProxyType({}),
#             ),
#             (
#                     errors.AttributesWithStaticDefaultsError,
#                     frozenset(),
#                     types.MappingProxyType(dict(name=str, age=int)),
#                     frozenset(("name", "age")),
#                     frozenset(("age",)),
#                     types.MappingProxyType({}),
#             ),
#             (
#                     errors.AttributesWithDynamicDefaultsError,
#                     frozenset(),
#                     types.MappingProxyType(dict(name=str, age=int)),
#                     frozenset(("name", "age")),
#                     frozenset(("age",)),
#                     types.MappingProxyType({}),
#             ),
#             (
#                     errors.ClassVarsError,
#                     frozenset(),
#                     types.MappingProxyType({}),
#                     frozenset(),
#                     frozenset(),
#                     types.MappingProxyType(dict(name="Username", age=42)),
#             ),
#             (
#                     errors.MixedError,
#                     frozenset(("name", "age", "note")),
#                     types.MappingProxyType(dict(name=str,
#                                                 age=int,
#                                                 timestamp=datetime.datetime,
#                                                 my_type=str)),
#                     frozenset(("name", "age", "note", "timestamp", "my_type")),
#                     frozenset(("age", "timestamp", "my_type")),
#                     types.MappingProxyType(dict(entity="The Entity")),
#             ),
#             (
#                     errors.DerivedError,
#                     frozenset(("name", "surname", "age", "note")),
#                     types.MappingProxyType(dict(name=str,
#                                                 age=int,
#                                                 timestamp=datetime.datetime,
#                                                 my_type=str,
#                                                 surname=str,
#                                                 location=t.Tuple[float, float],
#                                                 updated_at=datetime.datetime,
#                                                 full_name=str,
#                                                 box=dict)),
#                     frozenset(("name",
#                                "age",
#                                "note",
#                                "timestamp",
#                                "my_type",
#                                "surname",
#                                "location",
#                                "updated_at",
#                                "full_name",
#                                "box")),
#                     frozenset(("age",
#                                "timestamp",
#                                "my_type",
#                                "location",
#                                "updated_at",
#                                "full_name")),
#                     types.MappingProxyType(dict(entity="The Entity")),
#             ),
#     )
# )
# def test_cls_store(kls, fields, hints, registered, defaults, consts):
#     """Validates store management from root.Error.__init_subclass__"""
#
#     store = getattr(kls, "_Error__cls_store")
#
#     assert type(store.fields) is type(fields)
#     assert store.fields == fields
#     assert type(store.inst_hints) is type(hints)
#     assert store.inst_hints == hints
#     assert type(store.registered) is type(registered)
#     assert store.registered == registered
#     assert type(store.defaults) is type(defaults)
#     assert store.defaults == defaults
#     assert type(store.consts) is type(consts)
#     assert store.consts == consts
#
#
# @pytest.mark.parametrize(
#     "features",
#     (root.Features.FORBID_NON_NAMED_FIELDS,
#      root.Features.NONE)
# )
# @mock.patch("izulu._utils.check_non_named_fields", return_value=0)
# def test_cls_validation(mocked_check, features):
#     """Validates feature checks from root.Error.__init_subclass__"""
#
#     type("Err", (errors.ClassVarsError,), {"__features__": features})
#
#     if features is root.Features.NONE:
#         mocked_check.assert_not_called()
#     else:
#         mocked_check.assert_called_once()



import pytest

from izulu import _utils
from tests import helpers as h


@pytest.mark.parametrize(
    ("store", "kws"),
    [
        (h._make_store(), tuple()),
        (h._make_store(fields=("name", "age")), ("name", "age")),
        (
            h._make_store(
                fields=("name", "age"), inst_hints=dict(name=str, age=int)
            ),
            ("name", "age"),
        ),
        (
            h._make_store(
                fields=("name", "age"),
                inst_hints=dict(name=str, age=int),
                defaults=("age",),
            ),
            ("name",),
        ),
        (
            h._make_store(
                fields=("name", "age", "ENTITY"),
                inst_hints=dict(name=str, age=int),
                defaults=("age",),
                consts=dict(ENTITY="THING"),
            ),
            ("name",),
        ),
        (
            h._make_store(
                fields=("name", "age"),
                inst_hints=dict(name=str, age=int),
                defaults=("name", "age"),
                consts=dict(ENTITY="THING"),
            ),
            tuple(),
        ),
        (
            h._make_store(
                fields=("name", "age"),
                inst_hints=dict(name=str, age=int),
                defaults=("age",),
                consts=dict(ENTITY="THING"),
            ),
            ("name",),
        ),
        (
            h._make_store(
                fields=("name", "ENTITY"),
                inst_hints=dict(name=str),
                consts=dict(ENTITY="THING"),
            ),
            ("name",),
        ),
        (
            h._make_store(
                fields=("name", "ENTITY"), consts=dict(ENTITY="THING")
            ),
            ("name",),
        ),
        (
            h._make_store(
                inst_hints=dict(name=str, age=int), defaults=dict(age=42)
            ),
            ("name",),
        ),
        (
            h._make_store(defaults=("age",), consts=dict(ENTITY="THING")),
            ("name",),
        ),
    ],
)
def test_check_missing_fields_ok(store, kws):
    _utils.check_missing_fields(store, frozenset(kws))


@pytest.mark.parametrize(
    ("store", "kws"),
    [
        (h._make_store(fields=("name", "age")), tuple()),
        (h._make_store(fields=("name", "age")), ("age",)),
        (
            h._make_store(
                fields=("name",), inst_hints=dict(name=str, age=int)
            ),
            ("name",),
        ),
        (
            h._make_store(fields=("name", "age", "ENTITY"), defaults=("age",)),
            ("name",),
        ),
        (h._make_store(inst_hints=dict(name=str, age=int)), ("name",)),
        (
            h._make_store(
                fields=("name", "ENTITY"),
                inst_hints=dict(name=str),
                consts=dict(ENTITY="THING"),
            ),
            ("ENTITY",),
        ),
        (
            h._make_store(
                fields=("name", "ENTITY"), consts=dict(ENTITY="THING")
            ),
            ("age",),
        ),
    ],
)
def test_check_missing_fields_fail(store, kws):
    with pytest.raises(TypeError):
        _utils.check_missing_fields(store, frozenset(kws))


@pytest.mark.parametrize(
    ("store", "kws"),
    [
        (h._make_store(), tuple()),
        (h._make_store(fields=("name", "age")), ("name", "age")),
        (h._make_store(inst_hints=dict(name=str, age=int)), ("name", "age")),
        (h._make_store(const_hints=dict(name=str, age=int)), ("name", "age")),
        (
            h._make_store(
                fields=("name", "age", "ENTITY"),
                inst_hints=dict(name=str, age=int),
                defaults=("age",),
                const_hints=dict(ENTITY=str),
            ),
            ("name",),
        ),
        (
            h._make_store(
                fields=("name",),
                inst_hints=dict(name=str, age=int),
                defaults=("name", "age"),
                const_hints=dict(ENTITY=str),
            ),
            tuple(),
        ),
    ],
)
def test_check_undeclared_fields_ok(store, kws):
    _utils.check_undeclared_fields(store, frozenset(kws))


@pytest.mark.parametrize(
    ("store", "kws"),
    [
        (h._make_store(), ("entity",)),
        (h._make_store(fields=("name", "age")), ("entity",)),
        (h._make_store(inst_hints=dict(name=str, age=int)), ("entity",)),
        (h._make_store(consts=dict(name="John", age=42)), ("entity",)),
    ],
)
def test_check_undeclared_fields_fail(store, kws):
    with pytest.raises(TypeError):
        _utils.check_undeclared_fields(store, frozenset(kws))


@pytest.mark.parametrize(
    ("store", "kws"),
    [
        (h._make_store(), tuple()),
        (h._make_store(const_hints=dict(ENTITY=str)), tuple()),
        (h._make_store(const_hints=dict(ENTITY=str)), ("name",)),
    ],
)
def test_check_kwarg_consts_ok(store, kws):
    _utils.check_kwarg_consts(store, frozenset(kws))


@pytest.mark.parametrize(
    ("store", "kws"),
    [
        (h._make_store(const_hints=dict(ENTITY=str)), ("ENTITY",)),
        (h._make_store(const_hints=dict(ENTITY=str)), ("ENTITY", "name")),
    ],
)
def test_check_kwarg_consts_fail(store, kws):
    with pytest.raises(TypeError):
        _utils.check_kwarg_consts(store, frozenset(kws))


@pytest.mark.parametrize(
    "store",
    [
        h._make_store(),
        h._make_store(fields=frozenset(("abc", "-", "01abc"))),
    ],
)
def test_check_non_named_fields_ok(store):
    _utils.check_non_named_fields(store)


@pytest.mark.parametrize(
    "store",
    [
        h._make_store(fields=frozenset(("",))),
        h._make_store(fields=frozenset((1,))),
        h._make_store(fields=frozenset((0,))),
    ],
)
def test_check_non_named_fields_fail(store):
    with pytest.raises(ValueError, match="Field names can't be"):
        _utils.check_non_named_fields(store)



import datetime

import pytest

from tests import errors


@pytest.fixture
def derived_error():
    ts = datetime.datetime.now(datetime.timezone.utc)
    return errors.DerivedError(
        name="John",
        surname="Brown",
        note="...",
        age=42,
        updated_at=ts,
        full_name="secret",
        box=dict(),
    )



# import pytest
#
#
# class TestSomeModuleTestCase:
#
#     @pytest.fixture(autouse=True)
#     def context(self):
#         print("SETUP")
#         try:
#             yield
#         finally:
#             print("TEARDOWN")
#
#     def test_one(self):
#         print("    one")
#         assert True
#         print("    one after")
#
#     @pytest.mark.skip()
#     def test_two(self):
#         print("    two")
#         assert False
#         print("    two after")



import contextlib
import logging
import typing as t

_LOG = logging.getLogger(__name__)


@contextlib.contextmanager
def suppress(
    *excs: t.Type[Exception],
    exclude: t.Optional[t.Type[Exception]] = None,
) -> t.Generator[None, None, None]:
    exc_targets = excs or Exception
    try:
        yield
    except exc_targets as e:
        if exclude and isinstance(e, exclude):
            raise
        _LOG.error("Error suppressed: %s", e)



from unittest import mock

import pytest

from izulu import root
from tests import errors


@pytest.mark.parametrize(
    "kls",
    [errors.TemplateOnlyError, errors.AttributesOnlyError],
)
@pytest.mark.parametrize("kwargs", [dict(), dict(name="John"), dict(age=42)])
def test_forbid_missing_fields_triggered(kls, kwargs):
    toggles = {"__toggles__": root.Toggles.FORBID_MISSING_FIELDS}

    with pytest.raises(TypeError):
        type("TestError", (kls,), toggles)(**kwargs)


@pytest.mark.parametrize(
    ("kls", "kwargs"),
    [
        (errors.RootError, dict(field="value")),
        (errors.TemplateOnlyError, dict(name="John", age=42, field="field")),
        (errors.AttributesOnlyError, dict(name="John", age=42, field="field")),
    ],
)
def test_forbid_undeclared_fields_triggered(kls, kwargs):
    toggles = {"__toggles__": root.Toggles.FORBID_UNDECLARED_FIELDS}

    with pytest.raises(TypeError):
        type("TestError", (kls,), toggles)(**kwargs)


@pytest.mark.parametrize(
    ("kls", "kwargs"),
    [
        (errors.ClassVarsError, dict(name="John")),
        (errors.ClassVarsError, dict(age=0)),
        (errors.ClassVarsError, dict(blah=1.0)),
        (errors.ClassVarsError, dict(name="John", age=0)),
        (errors.ClassVarsError, dict(age=0, blah=1.0)),
        (errors.MixedError, dict(name="John", age=42, entity="thing")),
    ],
)
def test_forbid_kwarg_consts_triggered(kls, kwargs):
    toggles = {"__toggles__": root.Toggles.FORBID_KWARG_CONSTS}

    with pytest.raises(TypeError):
        type("TestError", (kls,), toggles)(**kwargs)


@pytest.mark.parametrize("toggles", [root.Toggles(i) for i in range(7)])
@mock.patch("izulu._utils.check_kwarg_consts")
@mock.patch("izulu._utils.check_undeclared_fields")
@mock.patch("izulu._utils.check_missing_fields")
def test_process_toggles(mock_missing, mock_undeclared, mock_const, toggles):
    with mock.patch.object(
        errors.RootError,
        "__toggles__",
        new_callable=mock.PropertyMock,
    ) as mocked:
        mocked.return_value = toggles

        e = errors.RootError()
        args = (e._Error__cls_store, frozenset(e._Error__kwargs))

        if root.Toggles.FORBID_MISSING_FIELDS in toggles:
            mock_missing.assert_called_once_with(*args)
        else:
            mock_missing.assert_not_called()
        if root.Toggles.FORBID_UNDECLARED_FIELDS in toggles:
            mock_undeclared.assert_called_once_with(*args)
        else:
            mock_undeclared.assert_not_called()
        if root.Toggles.FORBID_KWARG_CONSTS in toggles:
            mock_const.assert_called_once_with(*args)
        else:
            mock_const.assert_not_called()


def test_feature_presets():
    default = (
        root.Toggles.FORBID_MISSING_FIELDS
        | root.Toggles.FORBID_UNDECLARED_FIELDS
        | root.Toggles.FORBID_KWARG_CONSTS
        | root.Toggles.FORBID_NON_NAMED_FIELDS
        | root.Toggles.FORBID_UNANNOTATED_FIELDS
    )

    assert root.Toggles.NONE is root.Toggles(0)
    assert root.Toggles(0) == root.Toggles.NONE
    assert root.Toggles.DEFAULT is default
    assert default == root.Toggles.DEFAULT


def test_default_toggles():
    assert root.Error.__toggles__ is root.Toggles.DEFAULT



import copy
import datetime
import pickle

import pytest

from izulu import root
from tests import errors

TS = datetime.datetime.now(datetime.timezone.utc)


@pytest.mark.parametrize(
    ("err", "expected"),
    [
        (errors.RootError(), dict()),
        (
            errors.MixedError(name="John", age=10, note="..."),
            dict(name="John", age=10, note="..."),
        ),
        (
            errors.DerivedError(
                name="John", surname="Brown", note="...", box={}
            ),
            dict(name="John", surname="Brown", note="...", box={}),
        ),
    ],
)
def test_as_kwargs(err, expected):
    kwargs = err.as_kwargs()
    assert kwargs == expected

    kwargs["item"] = "SURPRISE"
    assert kwargs != err._Error__kwargs

    assert id(kwargs) != id(err._Error__kwargs)


@pytest.mark.parametrize(
    ("err", "expected", "wide"),
    [
        (errors.RootError(), dict(), False),
        (errors.RootError(), dict(), True),
        (errors.ClassVarsError(), dict(), False),
        (errors.ClassVarsError(), dict(age=42, name="Username"), True),
        (
            errors.AttributesWithStaticDefaultsError(name="John"),
            dict(name="John", age=0),
            False,
        ),
        (
            errors.AttributesWithStaticDefaultsError(name="John"),
            dict(name="John", age=0),
            True,
        ),
        (
            errors.MixedError(name="John", age=10, note="...", timestamp=TS),
            dict(
                name="John",
                age=10,
                note="...",
                my_type="MixedError",
                timestamp=TS,
            ),
            False,
        ),
        (
            errors.MixedError(name="John", age=10, note="...", timestamp=TS),
            dict(
                name="John",
                age=10,
                note="...",
                my_type="MixedError",
                timestamp=TS,
                entity="The Entity",
            ),
            True,
        ),
        (
            errors.DerivedError(
                name="John",
                surname="Brown",
                note="...",
                box={},
                timestamp=TS,
                updated_at=TS,
            ),
            dict(
                name="John",
                surname="Brown",
                full_name="John Brown",
                note="...",
                age=0,
                box={},
                location=(50.3, 3.608),
                my_type="DerivedError",
                timestamp=TS,
                updated_at=TS,
            ),
            False,
        ),
        (
            errors.DerivedError(
                name="John",
                surname="Brown",
                note="...",
                box={},
                timestamp=TS,
                updated_at=TS,
            ),
            dict(
                name="John",
                surname="Brown",
                full_name="John Brown",
                note="...",
                age=0,
                box={},
                location=(50.3, 3.608),
                my_type="DerivedError",
                timestamp=TS,
                updated_at=TS,
                entity="The Entity",
            ),
            True,
        ),
    ],
)
def test_as_dict(err, expected, wide):
    data = err.as_dict(wide=wide)
    assert data == expected

    data["item"] = "SURPRISE"
    assert "key" not in err._Error__kwargs

    with pytest.raises(AttributeError):
        _ = data.key

    assert id(data) != id(err._Error__kwargs)


def test_as_dict_wide_override_const():
    toggles = {"__toggles__": root.Toggles.NONE}
    kls = type("Err", (errors.ClassVarsError,), toggles)

    err = kls(age=500)

    assert err.as_dict(wide=True) == dict(age=500, name="Username")


def _assert_copy_mutual(orig, cp):
    assert cp is not orig
    assert cp._Error__kwargs is not orig._Error__kwargs
    assert cp.as_dict() == orig.as_dict()
    assert cp.box == orig.box == dict()


def test_copy_shallow(derived_error):
    orig = derived_error
    shallow = copy.copy(orig)

    assert shallow.box is orig.box
    _assert_copy_mutual(orig, shallow)

    orig.box.update(a=11)

    assert shallow.box == dict(a=11)


def test_copy_deep(derived_error):
    orig = derived_error
    deep = copy.deepcopy(orig)

    assert deep is not orig.box
    _assert_copy_mutual(orig, deep)

    orig.box.update(a=11)

    assert deep.box == dict()


@pytest.mark.parametrize(
    "err",
    [
        errors.RootError(),
        errors.TemplateOnlyError(name="John", age=42),
        errors.AttributesOnlyError(name="John", age=42),
        errors.MixedError(name="John", note="..."),
        errors.DerivedError(name="John", surname="Brown", note="...", box={}),
    ],
)
def test_pickling(err):
    dumped = pickle.dumps(err)
    resurrected = pickle.loads(dumped)

    assert str(err) == str(resurrected)
    assert repr(err) == repr(resurrected)
    assert err.as_dict() == resurrected.as_dict()



from __future__ import annotations

import contextlib
import logging
import typing as t

from izulu import _utils

_IMPORT_ERROR_TEXTS = (
    "",
    "You have early version of Python.",
    "  Extra compatibility dependency required.",
    "  Please add 'izulu[compatibility]' to your project dependencies.",
    "",
    "Pip: `pip install izulu[compatibility]`",
)


if hasattr(t, "dataclass_transform"):
    t_ext = t
else:
    try:
        import typing_extensions as t_ext  # type: ignore[no-redef]
    except ImportError:
        for message in _IMPORT_ERROR_TEXTS:
            logging.error(message)  # noqa: LOG015,TRY400
        raise

_T_KWARGS = t.Dict[str, t.Any]
_T_EXC_CLASS_OR_TUPLE = t.Union[
    t.Type[Exception],
    t.Tuple[t.Type[Exception], ...],
]
_T_FACTORY = t.Callable[
    [t.Type[Exception], Exception, _T_KWARGS],
    t.Optional[Exception],
]
_T_ACTION = t.Union[str, t.Type[Exception], _T_FACTORY, None]
_T_RULE = t.Tuple[t.Tuple[_T_EXC_CLASS_OR_TUPLE, _T_ACTION], ...]
_T_RULES = t.Union[
    bool,
    _T_RULE,  # tup, chain?
]
_T_RERAISING = t.Union[
    _T_RULE,  # tup, chain?
    bool,
    None,
]
_T_COMPILED_ACTION = t.Callable[[Exception, _T_KWARGS], t.Optional[Exception]]
_T_COMPILED_RULES = t.Union[
    bool,
    t.Tuple[t.Tuple[_T_EXC_CLASS_OR_TUPLE, _T_COMPILED_ACTION], ...],
]

_MISSING = object()

DecParam = t_ext.ParamSpec("DecParam")
DecReturnType = t.TypeVar("DecReturnType")


class FatalMixin:
    """
    Mark exception as non-recoverable.

    Should be directly inherited. You can't inherit from fatal exception.
    Fatal exceptions are by-passed by ``ReraisingMixin`` tools.
    """

    def __init_subclass__(cls, **kwargs: t.Any) -> None:  # noqa: ANN401
        if FatalMixin not in cls.__bases__:
            raise TypeError("Fatal can't be indirectly inherited")
        super().__init_subclass__(**kwargs)


class ReraisingMixin:
    __reraising__: _T_RULES = False

    __reraising: _T_COMPILED_RULES

    def __init_subclass__(cls, **kwargs: t.Any) -> None:  # noqa: ANN401
        super().__init_subclass__(**kwargs)
        rules = cls.__dict__.get("__reraising__", False)
        cls.__reraising = cls.__compile_rules(rules)

    @classmethod
    def __compile_rules(cls, rules: _T_RULES) -> _T_COMPILED_RULES:
        if isinstance(rules, bool):
            return rules

        return tuple(
            (exc_type, cls.__compile_action(action))
            for exc_type, action in rules
        )

    @classmethod
    def __compile_action(  # noqa: C901
        cls,
        action: _T_ACTION,
    ) -> _T_COMPILED_ACTION:
        if action is None:

            def compiled_action(
                orig: Exception,  # noqa: ARG001
                kwargs: _T_KWARGS,  # noqa: ARG001
            ) -> t.Optional[Exception]:
                return None

        # TODO(d.burmistrov): temporary ignore
        elif action is t_ext.Self:  # type: ignore[comparison-overlap]

            def compiled_action(
                orig: Exception,  # noqa: ARG001
                kwargs: _T_KWARGS,
            ) -> t.Optional[Exception]:
                kls = t.cast("t.Type[Exception]", cls)
                return kls(**kwargs)

        elif isinstance(action, str):

            def compiled_action(
                orig: Exception,
                kwargs: _T_KWARGS,
            ) -> t.Optional[Exception]:
                action_ = t.cast(
                    "t.Callable[[Exception, _T_KWARGS], t.Optional[Exception]]",  # noqa: E501
                    getattr(cls, action),
                )
                return action_(orig, kwargs)

        elif isinstance(action, type) and issubclass(action, Exception):

            def compiled_action(
                orig: Exception,  # noqa: ARG001
                kwargs: _T_KWARGS,
            ) -> t.Optional[Exception]:
                return action(**kwargs)

        elif callable(action):

            def compiled_action(
                orig: Exception,
                kwargs: _T_KWARGS,
            ) -> t.Optional[Exception]:
                kls = t.cast("t.Type[Exception]", cls)
                return t.cast("_T_FACTORY", action)(kls, orig, kwargs)

        else:
            raise ValueError(f"Unsupported action: {action}")

        return compiled_action

    @classmethod
    def remap(
        cls,
        exc: Exception,
        *,
        reraising: _T_RERAISING = None,
        remap_kwargs: t.Optional[_T_KWARGS] = None,
        original_over_none: bool = False,
    ) -> t.Union[Exception, None]:
        """
        Return remapped exception instance.

        Remapping rules:

        1. if the result of remapping is to leave the original exception
           method will return

           a. ``None`` for ``original_over_none=False``,
           b. original exception for ``original_over_none=True``.

        2. early-return rule works first to abort remapping process for:

           a. exception with ``FatalMixin``,
           b. descendant exceptions for used class.

        3. Default behaviour is not to remap exception.

        4. Rules: ...

        Args:
            exc: original exception to be remapped
            reraising: manual overriding reraising rules
            remap_kwargs: provide kwargs for reraise exception
            original_over_none: if ``True`` return original
                                exception instead of ``None``

        Returns:
            reraising context manager

        """
        reraising_ = cls.__reraising
        if reraising is not None:
            reraising_ = cls.__compile_rules(reraising)

        if (
            isinstance(exc, cls)
            or not reraising_
            or FatalMixin in exc.__class__.__bases__
        ):
            if original_over_none:
                return exc
            return None

        remap_kwargs = remap_kwargs or {}

        # greedy remapping (any occurred exception)
        if reraising_ is True:
            kls = t.cast("t.Type[Exception]", cls)
            return kls(**remap_kwargs)

        reraising__ = t.cast(
            "t.Tuple[t.Tuple[_T_EXC_CLASS_OR_TUPLE, _T_COMPILED_ACTION], ...]",
            reraising_,
        )

        for match, rule in reraising__:
            if not isinstance(exc, match):
                continue

            e = rule(exc, remap_kwargs)
            if e is None:
                break

            return e

        if original_over_none:
            return exc
        return None

    @classmethod
    @contextlib.contextmanager
    def reraise(  # type: ignore[no-untyped-def] # noqa: ANN206
        cls,
        reraising: _T_RERAISING = None,
        remap_kwargs: t.Optional[_T_KWARGS] = None,
    ):
        """
        Context Manager & Decorator to raise class exception over original.

        Args:
            reraising: manual overriding reraising rules
            remap_kwargs: provide kwargs for reraise exception

        """
        try:
            yield
        except Exception as e:  # noqa: BLE001
            orig = e
        else:
            return

        exc = cls.remap(
            exc=orig,
            reraising=reraising,
            remap_kwargs=remap_kwargs,
        )
        if exc is None:
            raise  # noqa: PLE0704

        raise exc from orig

    @classmethod
    @contextlib.asynccontextmanager
    async def async_reraise(  # type: ignore[no-untyped-def] # noqa: ANN206
        cls,
        reraising: _T_RERAISING = None,
        remap_kwargs: t.Optional[_T_KWARGS] = None,
    ):
        """
        Async version of `reraise`.

        Args:
            reraising: manual overriding reraising rules
            remap_kwargs: provide kwargs for reraise exception

        """
        with cls.reraise(reraising=reraising, remap_kwargs=remap_kwargs):
            yield


def skip(target: t.Type[Exception]) -> _T_RULE:
    return ((target, None),)


def catch(
    target: t.Type[Exception] = Exception,
    *,
    exclude: t.Optional[t.Type[Exception]] = None,
    new: t.Any = t_ext.Self,  # noqa: ANN401
) -> _T_RULE:
    rule = (target, new)
    if exclude:
        return (exclude, None), rule
    return (rule,)


class chain:  # noqa: N801
    def __init__(self, kls: ReraisingMixin, *klasses: ReraisingMixin) -> None:
        self._klasses = (kls, *klasses)

    def __call__(
        self,
        actor: t.Type[Exception],  # noqa: ARG002
        exc: Exception,
        reraising: _T_RERAISING = None,  # noqa: ARG002
        remap_kwargs: t.Optional[_T_KWARGS] = None,
    ) -> t.Optional[Exception]:
        for kls in self._klasses:
            remapped = kls.remap(exc=exc, remap_kwargs=remap_kwargs)
            if remapped is not None:
                return remapped
        return None

    @classmethod
    def from_subtree(cls, klass: t.Type[ReraisingMixin]) -> chain:
        it = (
            t.cast("ReraisingMixin", kls)
            for kls in _utils.traverse_tree(klass)
        )
        return cls(*it)

    @classmethod
    def from_names(cls, name: str, *names: str) -> chain:
        objects = globals()
        err_klasses = []
        for name in (name, *names):  # noqa: B020,PLR1704
            kls = objects.get(name, _MISSING)
            if kls is _MISSING:
                msg = f"module '{__name__}' has no attribute '{name}'"
                raise AttributeError(msg)

            err_klasses.append(kls)

        return cls(*err_klasses)



from __future__ import annotations

import collections
import copy
import functools
import types
import typing as t

from izulu import _utils


class BaseLogEvent(collections.UserString):
    __template__: t.ClassVar[str] = "Unspecified error"

    __cls_store: t.ClassVar[_utils.Store] = _utils.Store(
        fields=frozenset(),
        const_hints=types.MappingProxyType(dict()),
        inst_hints=types.MappingProxyType(dict()),
        consts=types.MappingProxyType(dict()),
        defaults=frozenset(),
    )

    def __init_subclass__(cls, **kwargs: t.Any) -> None:
        super().__init_subclass__(**kwargs)
        fields = frozenset(_utils.iter_fields(cls.__template__))
        const_hints, inst_hints = _utils.split_cls_hints(cls)
        consts = _utils.get_cls_defaults(cls, const_hints)
        defaults = _utils.get_cls_defaults(cls, inst_hints)
        cls.__cls_store = _utils.Store(
            fields=fields,
            const_hints=types.MappingProxyType(const_hints),
            inst_hints=types.MappingProxyType(inst_hints),
            consts=types.MappingProxyType(consts),
            defaults=frozenset(defaults),
        )

    def __init__(self, **kwargs: t.Any) -> None:
        self.__kwargs = kwargs.copy()
        self.__populate_attrs()
        self.data = self.__class__.__name__

    def __populate_attrs(self) -> None:
        """Set hinted kwargs as exception attributes"""
        for k, v in self.__kwargs.items():
            if k in self.__cls_store.inst_hints:
                setattr(self, k, v)

    def __process_template(self, data: dict[str, t.Any]) -> str:
        """Format the error template from provided data (kwargs & defaults)"""
        kwargs = self.__cls_store.consts.copy()
        kwargs.update(data)
        return _utils.format_template(self.__template__, kwargs)

    def _hook(
        self, store: _utils.Store, kwargs: dict[str, t.Any], msg: str
    ) -> str:
        return msg

    def __repr__(self) -> str:
        kwargs = _utils.join_kwargs(**self.as_dict())
        return f"{self.__module__}.{self.__class__.__qualname__}({kwargs})"

    def __copy__(self):
        return type(self)(**self.as_dict())

    def __deepcopy__(self, memo: dict[int, t.Any]):
        _id = id(self)
        if _id not in memo:
            kwargs = {
                k: copy.deepcopy(v, memo) for k, v in self.as_dict().items()
            }
            memo[_id] = type(self)(**kwargs)
        return memo[_id]

    def __reduce__(self) -> tuple[t.Any, ...]:
        return functools.partial(self.__class__, **self.as_dict()), tuple()

    def as_str(self) -> str:
        """Represent error as exception type with message"""
        return f"{self.__class__.__qualname__}: {self}"

    def as_kwargs(self) -> dict[str, t.Any]:
        """Return the copy of original kwargs used to initialize the error"""
        return self.__kwargs.copy()

    def as_dict(self, wide: bool = False) -> dict[str, t.Any]:
        """
        Represent error as dict of fields including default values

        By default, only *instance* data and defaults are provided.

        :param bool wide: if `True` *class* defaults will be included in result
        """
        d = self.__kwargs.copy()
        for field in self.__cls_store.defaults:
            d.setdefault(field, getattr(self, field))
        if wide:
            for field, const in self.__cls_store.consts.items():
                d.setdefault(field, const)
        return d



import uuid
from unittest import mock  # noqa: RUF100

import pytest

from izulu import root


@pytest.mark.parametrize("flag", [True, False])
def test_factory(flag):
    expected = uuid.uuid4()
    m = mock.Mock(return_value=expected)
    attr = root.factory(default_factory=m, self=flag)
    k = type("Klass", tuple(), {"attr_with_self": attr})()
    call_args = (k,)

    result = k.attr_with_self

    assert result is expected
    m.assert_called_once_with(*call_args[:flag])
